package h08;


import h08.Exceptions.FlightNotFoundException;

import java.util.Arrays;

/**
 * Represents an airport. An airport manages departing and arriving flights, allowing for their addition, removal, and retrieval based on the airport code.
 */
public class Airport {

    /**
     * The code of the airport.
     */
    private String airportCode;

    /**
     * The departing flights of the airport.
     */
    private Flight[] departingFlights;

    /**
     * The arriving flights to the airport.
     */
    private Flight[] arrivingFlights;

    /**
     * The number of departing flights of the airport.
     */
    private int departingSize;

    /**
     * The number of arriving flights to the airport.
     */
    private int arrivingSize;

    /**
     * Constructs a new airport with the specified airport code and initial capacity.
     *
     * @param airportCode     the code of the airport
     * @param initialCapacity the initial capacity of the airport
     */
    public Airport(String airportCode, int initialCapacity) {
        this.airportCode = airportCode;
        this.departingFlights = new Flight[initialCapacity];
        this.arrivingFlights = new Flight[initialCapacity];
        this.departingSize = 0;
        this.arrivingSize = 0;
    }

    /**
     * Adds a flight.
     *
     * @param flight      the flight to add
     * @param isDeparting if true, adds to departing flights, otherwise to arriving flights
     * @throws IllegalArgumentException if the flight's airport code doesn't match the airport's code
     */
    public void addFlight(Flight flight, boolean isDeparting) {
        //TODO H8.4.1
        if (isDeparting) {
            addDeparture(flight);
        } else {
            addArrival(flight);
        }
    }

    public void addDeparture(Flight flight) {
        if (!flight.getDeparture().equals(airportCode)) {
            throw new IllegalArgumentException("Flight's departure airport code does not match this airport's code");
        }

        if (departingFlights.length >= departingSize) {
            departingFlights = Arrays.copyOf(departingFlights, departingSize*2);
        }

        departingFlights[departingSize++] = flight;
    }

    public void addArrival(Flight flight) {
        if (!flight.getDestination().equals(airportCode)) {
            throw new IllegalArgumentException("Flight's arrival airport code does not match this airport's code");
        }

        if (arrivingFlights.length >= arrivingSize) {
            arrivingFlights = Arrays.copyOf(arrivingFlights, arrivingSize*2);
        }

        arrivingFlights[arrivingSize++] = flight;
    }

    /**
     * Removes a flight by flight number.
     *
     * @param flightNumber the flight number
     * @param isDeparting  if true, removes from departing flights, otherwise from arriving flights
     * @throws FlightNotFoundException if the flight is not found
     */
    public void removeFlight(String flightNumber, boolean isDeparting) throws FlightNotFoundException {
        //TODO H8.4.2
        if (isDeparting) {
            removeDeparture(flightNumber);
        } else {
            removeArrival(flightNumber);
        }
    }

    public void removeArrival(String flightNumber) throws FlightNotFoundException {
        if (removeFlight(arrivingSize, arrivingFlights, flightNumber)) {
            arrivingSize--;
            return;
        };

        throw new FlightNotFoundException("Arriving flight not found: " + flightNumber);
    }

    public void removeDeparture(String flightNumber) throws FlightNotFoundException {
        if (removeFlight(departingSize, departingFlights, flightNumber)) {
            departingSize--;
            return;
        }

        throw new FlightNotFoundException("Departing flight not found: " + flightNumber);
    }

    private boolean removeFlight(int size, Flight[] flights, String flightNumber) {
        int index = 0;
        boolean found = false;

        for (; index < size; index++) {
            if (flights[index].getFlightNumber().equals(flightNumber)) {
                found = true;
                break;
            }
        }

        System.out.println(index);

        if (found) {
            for (int i = index; i < size - 1; i++) {
                flights[i] = flights[i + 1];
            }

            flights[size-1] = null;
        }
        return found;
    }

    /**
     * Returns a flight by flight number.
     *
     * @param flightNumber the flight number
     * @param isDeparting  if true, searches in departing flights, otherwise in arriving flights
     * @return the flight with the specified flight number
     * @throws FlightNotFoundException if the flight is not found
     */
    public Flight getFlight(String flightNumber, boolean isDeparting) throws FlightNotFoundException {
        //TODO H8.4.3
        if (isDeparting) {
            return getDeparture(flightNumber);
        } else {
            return getArrival(flightNumber);
        }
    }

    private Flight getArrival(String flightNumber) throws FlightNotFoundException {
        for (int i = 0; i < arrivingSize; i++) {
            if (arrivingFlights[i].getFlightNumber().equals(flightNumber)) {
                return arrivingFlights[i];
            }
        }

        throw new FlightNotFoundException("Arriving flight not found: " + flightNumber);
    }

    private Flight getDeparture(String flightNumber) throws FlightNotFoundException {
        for (int i = 0; i < departingSize; i++) {
            if (departingFlights[i].getFlightNumber().equals(flightNumber)) {
                return departingFlights[i];
            }
        }

        throw new FlightNotFoundException("Departing flight not found: " + flightNumber);
    }

    /**
     * Returns the departing flights of the airport.
     *
     * @return the departing flights of the airport
     */
    public Flight[] getAllDepartingFlights() {
        return Arrays.copyOf(departingFlights, departingSize);
    }

    /**
     * Returns the arriving flights to the airport.
     *
     * @return the arriving flights to the airport
     */
    public Flight[] getAllArrivingFlights() {
        return Arrays.copyOf(arrivingFlights, arrivingSize);
    }

    /**
     * Returns the airport code.
     *
     * @return the airport code
     */
    public String getAirportCode() {
        return airportCode;
    }
}
